"""Exec module for managing Networks."""
from typing import Any
from typing import Dict

from idem_gcp.tool.gcp.generate.exec_context import ExecutionContext
from idem_gcp.tool.gcp.utils import get_project_from_account

__func_alias__ = {"list_": "list"}


async def list_(
    hub,
    ctx,
    project: str = None,
    filter: str = None,
    max_results: int = 500,
    order_by: str = None,
    page_token: str = None,
    return_partial_success: bool = False,
) -> Dict[str, Any]:
    r"""Retrieves the list of networks available to the specified project.

    Args:
        project(str, Optional):
            Project ID for this request.

        filter(str, Optional):
            A filter expression that filters resources listed in the response. Most Compute resources support two types of filter expressions: expressions that support regular expressions and expressions that follow API improvement proposal AIP-160. If you want to use AIP-160, your expression must specify the field name, an operator, and the value that you want to use for filtering. The value must be a string, a number, or a boolean. The operator must be either `=`, `!=`, `>`, `<`, `<=`, `>=` or `:`. For example, if you are filtering Compute Engine instances, you can exclude instances named `example-instance` by specifying `name != example-instance`. The `:` operator can be used with string fields to match substrings. For non-string fields it is equivalent to the `=` operator. The `:*` comparison can be used to test whether a key has been defined. For example, to find all objects with `owner` label use: ``` labels.owner:* ``` You can also filter nested fields. For example, you could specify `scheduling.automaticRestart = false` to include instances only if they are not scheduled for automatic restarts. You can use filtering on nested fields to filter based on resource labels. To filter on multiple expressions, provide each separate expression within parentheses. For example: ``` (scheduling.automaticRestart = true) (cpuPlatform = \"Intel Skylake\") ``` By default, each expression is an `AND` expression. However, you can include `AND` and `OR` expressions explicitly. For example: ``` (cpuPlatform = \"Intel Skylake\") OR (cpuPlatform = \"Intel Broadwell\") AND (scheduling.automaticRestart = true) ``` If you want to use a regular expression, use the `eq` (equal) or `ne` (not equal) operator against a single un-parenthesized expression with or without quotes or against multiple parenthesized expressions. Examples: `fieldname eq unquoted literal` `fieldname eq 'single quoted literal'` `fieldname eq \"double quoted literal\"` `(fieldname1 eq literal) (fieldname2 ne \"literal\")` The literal value is interpreted as a regular expression using Google RE2 library syntax. The literal value must match the entire field. For example, to filter for instances that do not end with name "instance", you would use `name ne .*instance`.

        max_results(int, Optional):
            The maximum number of results per page that should be returned. If the number of available results is larger than `maxResults`, Compute Engine returns a `nextPageToken` that can be used to get the next page of results in subsequent list requests. Acceptable values are `0` to `500`, inclusive. (Default: `500`)

        order_by(str, Optional):
            Sorts list results by a certain order. By default, results are returned in alphanumerical order based on the resource name. You can also sort results in descending order based on the creation timestamp using `orderBy=\"creationTimestamp desc\"`. This sorts results based on the `creationTimestamp` field in reverse chronological order (newest result first). Use this to sort resources like operations so that the newest operation is returned first. Currently, only sorting by `name` or `creationTimestamp desc` is supported.

        page_token(str, Optional):
            Specifies a page token to use. Set `pageToken` to the `nextPageToken` returned by a previous list request to get the next page of results.

        return_partial_success(bool, Optional):
            Opt-in for partial success behavior which provides partial results in case of failure. The default value is false.

    Examples:
        .. code-block: sls

            random-name:
              exec.run:
              - path: gcp.compute.network.list
              - kwargs:
                  project: project-name
    """
    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }

    project = get_project_from_account(ctx, project)

    ret = await hub.exec.gcp_api.client.compute.network.list(
        ctx,
        project=project,
        filter=filter,
        maxResults=max_results,
        orderBy=order_by,
        pageToken=page_token,
        returnPartialSuccess=return_partial_success,
    )

    if not ret["result"]:
        result["comment"] += ret["comment"]
        result["result"] = False
        return result

    result["ret"] = ret["ret"]["items"]
    return result


async def get(
    hub, ctx, resource_id: str = None, project: str = None, name: str = None
) -> Dict[str, Any]:
    r"""Returns the specified network. Gets a list of available networks by making a list() request.

    Args:
        resource_id(str, Optional):
            An identifier of the resource in the provider.

        project(str, Optional):
            Project ID for this request.

        name(str, Optional):
            Name of the network to return.

    Examples:
        .. code-block: sls

            random-name:
              exec.run:
              - path: gcp.compute.network.get
              - kwargs:
                  project: project-name
                  name: network-name
    """
    result = {
        "comment": [],
        "ret": None,
        "result": True,
    }

    project = get_project_from_account(ctx, project)

    if resource_id:
        ret = await hub.exec.gcp_api.client.compute.network.get(
            ctx, resource_id=resource_id
        )
    elif project and name:
        ret = await hub.exec.gcp_api.client.compute.network.get(
            ctx, project=project, network=name
        )
    else:
        result["result"] = False
        result["comment"] = [
            f"gcp.compute.network#get(): {name} either resource_id or project and network should be specified."
        ]
        return result

    if not ret["result"]:
        result["comment"] += ret["comment"]
        result["result"] = False
        return result

    result["ret"] = ret["ret"]
    return result


async def get_effective_firewalls(
    hub,
    ctx,
    project: str = None,
    network: str = None,
    resource_id: str = None,
    request_id: str = None,
) -> Dict[str, Any]:
    r"""Returns the effective firewalls on a given network.

    Args:
        network(str):
            Name of the network for this request.

        project(str, Optional):
            Project ID for this request. Defaults to None.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        request_id(str, Optional):
            An optional request ID to identify requests. Specify a unique request ID so that if you must retry your request, the server will know to ignore the request if it has already been completed. For example, consider a situation where you make an initial request and the request times out. If you make the request again with the same request ID, the server can check if original operation with the same request ID was received, and if so, will ignore the second request. This prevents clients from accidentally creating duplicate commitments. The request ID must be a valid UUID with the exception that zero UUID is not supported ( 00000000-0000-0000-0000-000000000000).  Defaults to None.

    Examples:
        .. code-block: sls

            random-name:
              exec.run:
              - path: gcp.compute.network.get_effective_firewalls
              - kwargs:
                  project: project-name
                  network: network-name
    """
    project = get_project_from_account(ctx, project)

    execution_context = ExecutionContext(
        resource_type="compute.network",
        method_name="getEffectiveFirewalls",
        method_params={
            "ctx": ctx,
            "resource_id": resource_id,
            "project": project,
            "network": network,
        },
    )
    return await hub.tool.gcp.generate.generic_exec.execute(execution_context)


async def list_peering_routes(
    hub,
    ctx,
    peering_name: str,
    region: str,
    direction: str,
    network: str = None,
    resource_id: str = None,
    project: str = None,
    filter: str = None,
    order_by: str = None,
) -> Dict[str, Any]:
    r"""Lists the peering routes exchanged over peering connection.

    Args:
        peering_name(str):
            The response will show routes exchanged over the given peering connection.

        direction(str):
            The direction of the exchanged routes.
            Enum type. Allowed values:
            "INCOMING" - For routes exported from peer network.
            "OUTGOING" - For routes exported from local network.

        region(str):
            The region of the request. The response will include all subnet routes, static routes and dynamic routes in the region.

        project(str, Optional):
            Project ID for this request.

        network(str, Optional):
            Name of the network for this request.

        resource_id(str, Optional):
            An identifier of the resource in the provider.

        filter(str, Optional):
            A filter expression that filters resources listed in the response. Most Compute resources support two types of filter expressions: expressions that support regular expressions and expressions that follow API improvement proposal AIP-160. If you want to use AIP-160, your expression must specify the field name, an operator, and the value that you want to use for filtering. The value must be a string, a number, or a boolean. The operator must be either `=`, `!=`, `>`, `<`, `<=`, `>=` or `:`. For example, if you are filtering Compute Engine instances, you can exclude instances named `example-instance` by specifying `name != example-instance`. The `:` operator can be used with string fields to match substrings. For non-string fields it is equivalent to the `=` operator. The `:*` comparison can be used to test whether a key has been defined. For example, to find all objects with `owner` label use: ``` labels.owner:* ``` You can also filter nested fields. For example, you could specify `scheduling.automaticRestart = false` to include instances only if they are not scheduled for automatic restarts. You can use filtering on nested fields to filter based on resource labels. To filter on multiple expressions, provide each separate expression within parentheses. For example: ``` (scheduling.automaticRestart = true) (cpuPlatform = \"Intel Skylake\") ``` By default, each expression is an `AND` expression. However, you can include `AND` and `OR` expressions explicitly. For example: ``` (cpuPlatform = \"Intel Skylake\") OR (cpuPlatform = \"Intel Broadwell\") AND (scheduling.automaticRestart = true) ``` If you want to use a regular expression, use the `eq` (equal) or `ne` (not equal) operator against a single un-parenthesized expression with or without quotes or against multiple parenthesized expressions. Examples: `fieldname eq unquoted literal` `fieldname eq 'single quoted literal'` `fieldname eq \"double quoted literal\"` `(fieldname1 eq literal) (fieldname2 ne \"literal\")` The literal value is interpreted as a regular expression using Google RE2 library syntax. The literal value must match the entire field. For example, to filter for instances that do not end with name "instance", you would use `name ne .*instance`.

        order_by(str, Optional):
            Sorts list results by a certain order. By default, results are returned in alphanumerical order based on the resource name. You can also sort results in descending order based on the creation timestamp using `orderBy=\"creationTimestamp desc\"`. This sorts results based on the `creationTimestamp` field in reverse chronological order (newest result first). Use this to sort resources like operations so that the newest operation is returned first. Currently, only sorting by `name` or `creationTimestamp desc` is supported.

    Examples:
        .. code-block: sls

            random-name:
              exec.run:
              - path: gcp.compute.network.list_peering_routes
              - kwargs:
                  project: project-name
                  region: region-name
                  network: network-name
                  direction: direction-name
                  peering_name: peering-name
    """
    project = get_project_from_account(ctx, project)

    # GCP method name is simply the name of this func or the one from __func_alias - we can pass it
    # also we can pass the resource name based on the file name
    execution_context = ExecutionContext(
        resource_type="compute.network",
        method_name="listPeeringRoutes",
        method_params={
            "ctx": ctx,
            "network": network,
            "peering_name": peering_name,
            "region": region,
            "direction": direction,
            "resource_id": resource_id,
            "project": project,
            "filter": filter,
            "order_by": order_by,
        },
    )
    return await hub.tool.gcp.generate.generic_exec.execute(execution_context)


async def switch_to_custom_mode(
    hub,
    ctx,
    project: str = None,
    network: str = None,
    resource_id: str = None,
    request_id: str = None,
) -> Dict[str, Any]:
    r"""Switches the network mode from auto subnet mode to custom subnet mode.

    Args:
        network(str, Optional):
            Name of the network to be updated.

        project(str, Optional):
            Project ID for this request. Defaults to None.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        request_id(str, Optional):
            An optional request ID to identify requests. Specify a unique request ID so that if you must retry your request, the server will know to ignore the request if it has already been completed. For example, consider a situation where you make an initial request and the request times out. If you make the request again with the same request ID, the server can check if original operation with the same request ID was received, and if so, will ignore the second request. This prevents clients from accidentally creating duplicate commitments. The request ID must be a valid UUID with the exception that zero UUID is not supported ( 00000000-0000-0000-0000-000000000000).  Defaults to None.

    Examples:
        .. code-block: sls

            random-name:
              exec.run:
              - path: gcp.compute.network.switch_to_custom_mode
              - kwargs:
                  project: project-name
                  network: network-name
    """
    project = get_project_from_account(ctx, project)

    execution_context = ExecutionContext(
        resource_type="compute.network",
        method_name="switchToCustomMode",
        method_params={
            "ctx": ctx,
            "resource_id": resource_id,
            "project": project,
            "network": network,
        },
    )
    return await hub.tool.gcp.generate.generic_exec.execute(execution_context)



"""Exec module for managing Cloud Key Management Service crypto key versions."""
import base64
from typing import Any
from typing import Dict


__func_alias__ = {"list_": "list", "import_": "import"}


async def get(
    hub,
    ctx,
    resource_id: str,
):
    """Returns a crypto key version by its Idem resource ID.

    Args:
        resource_id(str):
            Idem resource ID. ``projects/{project id}/locations/{location id}/keyRings/{keyRing}/cryptoKeys/{cryptoKey}/cryptoKeyVersions/{cryptoKeyVersion}``

    Returns:
        CryptoKeyVersion resource

    Examples:
        .. code-block:: sls

            {% set project_id = 'project-name' %}
            {% set location_id = 'us-east1' %}
            {% set key_ring = 'key-ring' %}
            {% set crypto_key = 'crypto-key' %}
            {% set crypto_key_version = 'crypto-key-version' %}
            get-crypto-key-version:
                exec.run:
                    - path: gcp.cloudkms.crypto_key_version.get
                    - kwargs:
                        resource_id: projects/{{project_id}}/locations/{{location_id}}/keyRings/{{key_ring}}/cryptoKeys/{{crypto_key}}/cryptoKeyVersions/{{crypto_key_version}}
    """
    result = {
        "comment": [],
        "ret": [],
        "result": True,
    }

    crypto_key = await hub.exec.gcp_api.client.cloudkms.projects.locations.key_rings.crypto_keys.crypto_key_versions.get(
        ctx, _name=resource_id
    )

    if not crypto_key["result"]:
        result["comment"] += crypto_key["comment"]
        result["result"] = False
        return result

    result["ret"] = crypto_key["ret"]

    if not result["ret"]:
        result["comment"] += (
            hub.tool.gcp.comment_utils.get_empty_comment(
                "gcp.cloudkms.crypto_key_version", resource_id
            ),
        )

    return result


async def list_(
    hub, ctx, crypto_key: str, filter: str = None, order_by: str = None
) -> Dict[str, Any]:
    r"""Retrieves the crypto key versions in a crypto key.

    Args:
        crypto_key(str):
            key ring resource_id.

        filter(str, Optional):
            Only include resources that match the filter in the response. For more information, see
            `Sorting and filtering list results`_.

        order_by(str, Optional):
            Specify how the results should be sorted. If not specified, the results will be sorted in the default order.
            For more information, see `Sorting and filtering list results`_.

    .. _Sorting and filtering list results: https://cloud.google.com/kms/docs/sorting-and-filtering

    Examples:
        .. code-block:: sls

            list-crypto-key_versions-filtered:
                exec.run:
                   - path: gcp.cloudkms.crypto_key_version.list
                   - kwargs:
                         crypto_key: projects/project-name/locations/global/keyRings/kr-global-test
                         filter: algorithm = GOOGLE_SYMMETRIC_ENCRYPTION
    """
    result = {
        "comment": [],
        "ret": [],
        "result": True,
    }

    crypto_key_versions = await hub.exec.gcp_api.client.cloudkms.projects.locations.key_rings.crypto_keys.crypto_key_versions.list(
        ctx, parent=crypto_key, filter=filter, orderBy=order_by
    )
    if not crypto_key_versions["result"]:
        result["comment"] += crypto_key_versions["comment"]
        result["result"] = False
        return result

    if crypto_key_versions["ret"].get("items"):
        result["ret"].extend(crypto_key_versions["ret"]["items"])

    if not result["ret"]:
        result["comment"] += (
            hub.tool.gcp.comment_utils.list_empty_comment(
                "gcp.cloudkms.crypto_key_version", crypto_key
            ),
        )

    return result


async def import_(
    hub,
    ctx,
    parent: str,
    import_job: str,
    import_job_pub_key: str,
    algorithm: str,
    key_material: str,
    crypto_key_version: str = None,
) -> Dict[str, Any]:
    r"""Import key material in crypto key version.

    Args:
        parent:
            Required. The Idem resource_id of the `CryptoKey`_ to be imported into.
            The create permission is only required on this key when creating a new `CryptoKeyVersion`_.
            Authorization requires the following `IAM`_ permission on the specified resource parent:
                - cloudkms.cryptoKeyVersions.create
        import_job:
            Required. Idem resource_id of the import job to be used.
        import_job_pub_key:
            Required. PEM encoded public key of the import job to be used to wrap this key material.
            Authorization requires the following `IAM`_ permission on the specified resource import_job:
                - cloudkms.importjobs.useToImport
        algorithm:
            Required. The `algorithm`_ of the key being imported. This does not need to match the `versionTemplate`_ of the
            `CryptoKey`_ this version imports into.
        key_material:
            Base64 encoded key material. If importing symmetric key material, it is expected that the key contains
            plain bytes. If importing asymmetric key material, it is expected that the key is in
            PKCS#8-encoded DER format (the PrivateKeyInfo structure from RFC 5208).
        crypto_key_version(str, Optional):
            The optional Idem resource_id of an existing `CryptoKeyVersion`_ to target for an import operation. If this field is
            not present, a new `CryptoKeyVersion`_ containing the supplied key material is created.
            If this field is present, the supplied key material is imported into the existing `CryptoKeyVersion`_. To
            import into an existing `CryptoKeyVersion`_, the `CryptoKeyVersion`_ must be a child of
            `ImportCryptoKeyVersionRequest.parent`_, have been previously created via [cryptoKeyVersions.import][], and
            be in `DESTROYED`_ or `IMPORT_FAILED`_ state. The key material and algorithm must match the previous
            `CryptoKeyVersion`_ exactly if the `CryptoKeyVersion`_ has ever contained key material.
            Authorization requires the following `IAM`_ permission on the specified resource cryptoKeyVersion:
                - cloudkms.cryptoKeyVersions.update

    .. _CryptoKeyVersion: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.cryptoKeys.cryptoKeyVersions#CryptoKeyVersion
    .. _name: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.cryptoKeys.cryptoKeyVersions#CryptoKeyVersion.FIELDS.name
    .. _ImportCryptoKeyVersionRequest.parent: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.cryptoKeys.cryptoKeyVersions/import#body.PATH_PARAMETERS.parent
    .. _DESTROYED: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.cryptoKeys.cryptoKeyVersions#CryptoKeyVersion.CryptoKeyVersionState.ENUM_VALUES.DESTROYED
    .. _IMPORT_FAILED: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.cryptoKeys.cryptoKeyVersions#CryptoKeyVersion.CryptoKeyVersionState.ENUM_VALUES.IMPORT_FAILED
    .. _IAM: https://cloud.google.com/iam/docs/
    .. _CryptoKey: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.cryptoKeys#CryptoKey
    .. _algorithm: https://cloud.google.com/kms/docs/reference/rest/v1/CryptoKeyVersionAlgorithm
    .. _versionTemplate: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.cryptoKeys#CryptoKey.FIELDS.version_template
    .. _ImportJob: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.importJobs#ImportJob
    .. _RSA_OAEP_3072_SHA1_AES_256: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.importJobs#ImportMethod.ENUM_VALUES.RSA_OAEP_3072_SHA1_AES_256
    .. _RSA_OAEP_4096_SHA1_AES_256: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.importJobs#ImportMethod.ENUM_VALUES.RSA_OAEP_4096_SHA1_AES_256
    .. _public_key: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.importJobs#ImportJob.FIELDS.public_key

    Returns:
        Dict[str, Any] - If successful, the response body contains an instance of `CryptoKeyVersion`_.

    Examples:
        .. code-block:: sls

            {% set project_id = 'project-name' %}
            {% set location_id = 'us-east1' %}
            {% set key_ring_id = 'key-ring' %}
            {% set crypto_key_id = 'crypto-key' %}
            {% set import_job_id = 'import-job-id' %}

            import-job:
              gcp.cloudkms.import_job.present:
                  - import_method: RSA_OAEP_3072_SHA1_AES_256
                  - protection_level: SOFTWARE
                  - project_id: {{project_id}}
                  - location_id: {{location_id}}
                  - key_ring_id: {{key_ring_id}}
                  - import_job_id: {{import_job_id}}

            import-crypto-key:
                exec.run:
                   - path: gcp.cloudkms.crypto_key_version.import
                   - kwargs:
                         parent: projects/{{project_id}}/locations/{{location_id}}/keyRings/{{key_ring_id}}/cryptoKeys/{{crypto_key_id}}
                         import_job: ${gcp.cloudkms.import_job:import-job:resource_id}
                         import_job_pub_key: ${gcp.cloudkms.import_job:import-job:public_key:pem}
                         algorithm: "EC_SIGN_P256_SHA256"
                         key_material: rr5Y2UNi6+i3UQDrR8PO6s5ajAorN/SnHfZu+OCHx+w=
    """
    result = {
        "comment": [],
        "ret": [],
        "result": True,
    }

    rsa_aes_wrapped_key = hub.tool.gcp.cloudkms.crypto_key_version_utils.wrap_key(
        base64.b64decode(key_material), import_job_pub_key
    )

    import_ret = await hub.exec.gcp_api.client.cloudkms.projects.locations.key_rings.crypto_keys.crypto_key_versions.import_(
        ctx,
        parent=parent,
        body={
            "algorithm": algorithm,
            "importJob": import_job,
            "rsaAesWrappedKey": base64.urlsafe_b64encode(rsa_aes_wrapped_key).decode(
                "UTF-8"
            ),
            "cryptoKeyVersion": crypto_key_version,
        },
    )

    result["result"] = import_ret["result"]
    result["comment"] += import_ret["comment"]
    result["ret"] = import_ret["ret"]
    return result



from typing import Any
from typing import Dict

import deepdiff


def calc_update_mask(hub, sls_data: Dict[str, Any], gcp_data: Dict[str, Any]) -> str:
    """Calculate update mask needed for patch methods in cloudkms.

    Do not include in the update mask properties present only in gcp_data.

    Args:
        sls_data: First object to diff.
        gcp_data: Second object to diff

    Returns:
         List of fields to be updated in this request. None if there are no differences.

         This is a comma-separated list of fully qualified names of fields. Example: "user.displayName,photo".
    """
    if type(sls_data) != type(gcp_data):
        raise Exception(
            f"Dictionary params are of different types ({type(sls_data)} - {type(gcp_data)}."
            " deepdiff will not work correctly"
        )
    diff = deepdiff.DeepDiff(sls_data, gcp_data, ignore_order=True, view="tree")

    paths = set()
    for k in diff.affected_paths:
        if (
            isinstance(k.t1, deepdiff.helper.NotPresent)
            and not k.get_root_key() == "labels"
        ):
            continue
        p = k.path(output_format="list")
        # ["labels", "label_key"]
        root_key = k.get_root_key()
        if root_key == "labels":
            paths.add(root_key)
        else:
            paths.add(".".join(p))

    return ",".join(paths) if len(paths) > 0 else None


def merge_labels(
    hub, new_labels: Dict[str, str], old_labels: Dict[str, str]
) -> Dict[str, str]:
    if new_labels is None:
        return old_labels or {}
    elif old_labels is None:
        return new_labels

    return {**old_labels, **new_labels}



"""Google Cloud Platform provider exceptions module.

Copyright (c) 2021 VMware, Inc. All Rights Reserved.
SPDX-License-Identifier: Apache-2.0
"""


class NotFoundError(Exception):
    """Exception raised for resources not found in GCP."""

    def __init__(self, message: str):
        """Initialize an instance of this class.

        :param message: String message to present when rendering the exception.
        """
        self.message = message
        super().__init__(self.message)


class OperationError(Exception):
    """Exception raised for failed operations in GCP."""

    def __init__(self, message: str):
        """Initialize an instance of this class.

        :param message: String message to present when rendering the exception.
        """
        self.message = message
        super().__init__(self.message)



"""State module for managing Machine Images."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

from idem_gcp.tool.gcp.utils import get_project_from_account

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    project: str = None,
    resource_id: str = None,
    source_instance: str = None,
    storage_locations: List[str] = None,
    source_disk_encryption_keys: List[
        make_dataclass(
            "SourceDiskEncryptionKey",
            [
                ("sourceDisk", str, field(default=None)),
                ("diskEncryptionKey", {}, field(default=None)),
            ],
        )
    ] = None,
    description: str = None,
    saved_disks: List[
        make_dataclass(
            "SavedDisk",
            [
                ("kind", str, field(default="compute#savedDisk")),
                ("storageBytesStatus", str, field(default=None)),
                ("architecture", str, field(default=None)),
                ("storageBytes", str, field(default=None)),
                ("sourceDisk", str, field(default=None)),
            ],
        )
    ] = None,
    guest_flush: bool = None,
    machine_image_encryption_key: make_dataclass(
        "CustomerEncryptionKey",
        [
            ("kmsKeyServiceAccount", str, field(default=None)),
            ("rsaEncryptedKey", str, field(default=None)),
            ("rawKey", str, field(default=None)),
            ("kmsKeyName", str, field(default=None)),
        ],
    ) = None,
) -> Dict[str, Any]:
    r"""Creates a machine image in the specified project using the data that is included in the request. If you are creating a new machine image to update an existing instance, your new machine image should use the same network or, if applicable, the same subnetwork as the original instance.

    Args:

        name(str):
            An Idem name of the resource.

        project(str, Optional):
            Project ID for this request.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        source_instance(str, Optional):
            The source instance used to create the machine image. You can provide this as a partial or full URL to the resource. For example, the following are valid values: - https://www.googleapis.com/compute/v1/projects/project/zones/zone /instances/instance - projects/project/zones/zone/instances/instance . Defaults to None.

        storage_locations(list[str], Optional):
            The regional or multi-regional Cloud Storage bucket location where the machine image is stored. Defaults to None.

        source_disk_encryption_keys(list[Dict[str, Any]], Optional):
            [Input Only] The customer-supplied encryption key of the disks attached to the source instance. Required if the source disk is protected by a customer-supplied encryption key. Defaults to None.
            * properties (Dict[str, Dict[str, Any]], optional): None
                * sourceDisk (str, optional): URL of the disk attached to the source instance. This can be a full or valid partial URL. For example, the following are valid values: - https://www.googleapis.com/compute/v1/projects/project/zones/zone /disks/disk - projects/project/zones/zone/disks/disk - zones/zone/disks/disk
                * diskEncryptionKey ({}, optional): The customer-supplied encryption key of the source disk. Required if the source disk is protected by a customer-supplied encryption key.

        description(str, Optional):
            An optional description of this resource. Provide this property when you create the resource. Defaults to None.

        saved_disks(list[Dict[str, Any]], Optional):
            An array of Machine Image specific properties for disks attached to the source instance. Defaults to None.
            * properties (Dict[str, Dict[str, Any]], optional): An instance-attached disk resource.
                * kind (str, optional): [Output Only] Type of the resource. Always compute#savedDisk for attached disks.
                * storageBytesStatus (str, optional): [Output Only] An indicator whether storageBytes is in a stable state or it is being adjusted as a result of shared storage reallocation. This status can either be UPDATING, meaning the size of the snapshot is being updated, or UP_TO_DATE, meaning the size of the snapshot is up-to-date.
                * architecture (str, optional): [Output Only] The architecture of the attached disk.
                * storageBytes (str, optional): [Output Only] Size of the individual disk snapshot used by this machine image.
                * sourceDisk (str, optional): Specifies a URL of the disk attached to the source instance.

        guest_flush(bool, Optional):
            [Input Only] Whether to attempt an application consistent machine image by informing the OS to prepare for the snapshot process. Defaults to None.

        machine_image_encryption_key({}, Optional):
            Encrypts the machine image using a customer-supplied encryption key. After you encrypt a machine image using a customer-supplied key, you must provide the same key if you use the machine image later. For example, you must provide the encryption key when you create an instance from the encrypted machine image in a future request. Customer-supplied encryption keys do not protect access to metadata of the machine image. If you do not provide an encryption key when creating the machine image, then the machine image will be encrypted using an automatically generated key and you do not need to provide a key to use the machine image later. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            example_resource_name:
                gcp.compute.machine_image.present:
                  - project: project-name
                  - source_instance: https://www.googleapis.com/compute/v1/projects/project-name/zones/zone-name/instances/instance-name
    """
    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }

    project = get_project_from_account(ctx, project)

    # Get the resource_id from ESM
    if not resource_id:
        resource_id = (ctx.get("old_state") or {}).get(
            "resource_id"
        ) or hub.tool.gcp.resource_prop_utils.construct_resource_id(
            "compute.machine_image", {**locals(), "machineImage": name}
        )
    old = None

    # TODO: Handle operation result state
    if ctx.get("rerun_data"):
        handle_operation_ret = await hub.tool.gcp.operation_utils.handle_operation(
            ctx, ctx.get("rerun_data"), "compute.machine_image"
        )

        if not handle_operation_ret["result"]:
            result["result"] = False
            result["comment"] += handle_operation_ret["comment"]
            result["rerun_data"] = handle_operation_ret["rerun_data"]
            return result

        resource_id = handle_operation_ret["resource_id"]

    if resource_id:
        old_get_ret = await hub.exec.gcp_api.client.compute.machine_image.get(
            ctx, resource_id=resource_id
        )

        if not old_get_ret["result"]:
            # error other than non-existing resource
            result["result"] = False
            result["comment"] += old_get_ret["comment"]
            return result

        if old_get_ret["ret"]:
            # resource already exists

            # copy.copy(old_get_ret['ret']) is needed to convert any objects of type NamespaceDict to dict
            # in old_get_ret['ret']. This is done so that comparisons with the plan_state which has
            # objects of type dict works properly
            old = copy.deepcopy(copy.copy(old_get_ret["ret"]))
            result["old_state"] = old

            # If rerun_data is set at this point this means that there has been a completed create or update operation
            # so there is no need to run the code for create or update below again and we stop here
            if ctx.get("rerun_data"):
                result["new_state"] = result["old_state"]
                return result

    # TODO: Check if body contains the same parameters as old
    # to be autogenerated by pop-create based on insert/update props in properties.yaml
    resource_body = {
        "storage_locations": storage_locations,
        "guest_flush": guest_flush,
        "source_instance": source_instance,
        "description": description,
        "name": name,
        "saved_disks": saved_disks,
        "machine_image_encryption_key": machine_image_encryption_key,
        "source_disk_encryption_keys": source_disk_encryption_keys,
    }

    # TODO: How to handle query params which are not returned in the response body of get
    plan_state = {
        # "project": project,
        # "predefined_acl": predefined_acl,
        # "predefined_default_object_acl": predefined_default_object_acl,
        "resource_id": resource_id,
        **resource_body,
    }

    plan_state = {k: v for (k, v) in plan_state.items() if v is not None}
    operation = None
    if old:
        changes = hub.tool.gcp.utils.compare_states(
            old, plan_state, "compute.machine_image"
        )
        if not changes:
            result["result"] = True
            result["comment"].append(
                hub.tool.gcp.comment_utils.already_exists_comment(
                    "gcp.compute.machine_image", name
                )
            )
            result["new_state"] = result["old_state"]
            return result

        result["comment"].append(
            "No-op: There is no update function for gcp.compute.machine_image"
        )

    else:
        if ctx["test"]:
            result["comment"].append(
                hub.tool.gcp.comment_utils.would_create_comment(
                    "gcp.compute.machine_image", name
                )
            )
            result["new_state"] = plan_state
            return result

        # Create
        create_ret = await hub.exec.gcp_api.client.compute.machine_image.insert(
            ctx, project=project, body=resource_body
        )
        if not create_ret["result"]:
            result["result"] = False
            result["comment"] += create_ret["comment"]
            return result
        result["comment"].append(
            hub.tool.gcp.comment_utils.create_comment("gcp.compute.machine_image", name)
        )
        result["old_state"] = {}
        resource_id = create_ret["ret"].get("resource_id")
        if "compute#operation" in create_ret["ret"].get("kind"):
            operation = create_ret["ret"]

    if operation:
        operation_id = hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
            operation.get("selfLink"), "compute.global_operation"
        )
        handle_operation_ret = await hub.tool.gcp.operation_utils.handle_operation(
            ctx, operation_id, "compute.machine_image"
        )

        if not handle_operation_ret["result"]:
            result["result"] = False
            result["comment"] += handle_operation_ret["comment"]
            result["rerun_data"] = handle_operation_ret["rerun_data"]
            return result

        resource_id = handle_operation_ret["resource_id"]

    # Try getting the resource again
    # TODO: Reconciliation or waiter loop?
    # TODO: Check if this can be removed because insert and update may also return the necessary data on success
    get_ret = await hub.exec.gcp_api.client.compute.machine_image.get(
        ctx, resource_id=resource_id
    )

    if not get_ret["result"] and not get_ret["ret"]:
        result["result"] = False
        result["comment"] += get_ret["comment"]
        return result

    result["new_state"] = get_ret["ret"]

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    request_id: str = None,
) -> Dict[str, Any]:
    r"""Deletes the specified machine image. Deleting a machine image is permanent and cannot be undone.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        request_id(str, Optional):
            An optional request ID to identify requests. Specify a unique request ID so that if you must retry your request, the server will know to ignore the request if it has already been completed. For example, consider a situation where you make an initial request and the request times out. If you make the request again with the same request ID, the server can check if original operation with the same request ID was received, and if so, will ignore the second request. This prevents clients from accidentally creating duplicate commitments. The request ID must be a valid UUID with the exception that zero UUID is not supported ( 00000000-0000-0000-0000-000000000000). Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              gcp.compute.machine_image.absent:
                - name: value
                - project: value
                - machine_image: value
    """
    result = {
        "comment": [],
        "old_state": ctx.get("old_state"),
        "new_state": None,
        "name": name,
        "result": True,
    }

    if not resource_id:
        resource_id = (ctx.get("old_state") or {}).get("resource_id")

    if ctx.test:
        result["comment"].append(
            hub.tool.gcp.comment_utils.would_delete_comment(
                "gcp.compute.machine_image", name
            )
        )
        return result

    if not ctx.get("rerun_data"):
        # First iteration; invoke machine image's delete()
        delete_ret = await hub.exec.gcp_api.client.compute.machine_image.delete(
            ctx, resource_id=resource_id
        )
        if delete_ret["ret"]:
            if "compute#operation" in delete_ret["ret"].get("kind"):
                result["result"] = False
                result["comment"] += delete_ret["comment"]
                result[
                    "rerun_data"
                ] = hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
                    delete_ret["ret"].get("selfLink"), "compute.global_operation"
                )
                return result

    else:
        # delete() has been called on some previous iteration
        handle_operation_ret = await hub.tool.gcp.operation_utils.handle_operation(
            ctx, ctx.get("rerun_data"), "compute.machine_image"
        )
        if not handle_operation_ret["result"]:
            result["result"] = False
            result["comment"] += handle_operation_ret["comment"]
            result["rerun_data"] = handle_operation_ret["rerun_data"]
            return result

        resource_id = handle_operation_ret["resource_id"]

    if not resource_id:
        result["comment"].append(
            hub.tool.gcp.comment_utils.already_absent_comment(
                "gcp.compute.machine_image", name
            )
        )
        return result

    result["comment"].append(
        hub.tool.gcp.comment_utils.delete_comment("gcp.compute.machine_image", name)
    )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Retrieves a list of machine images that are contained within the specified project.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe gcp.compute.machine_image
    """
    result = {}

    describe_ret = await hub.exec.gcp_api.client.compute.machine_image.list(
        ctx, project=ctx.acct.project_id
    )

    if not describe_ret["result"]:
        hub.log.debug(
            f"Could not describe gcp.compute.machine_image {describe_ret['comment']}"
        )
        return {}

    for resource in describe_ret["ret"]["items"]:
        resource_id = resource.get("resource_id")
        result[resource_id] = {
            "gcp.compute.machine_image.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    return hub.tool.gcp.utils.is_pending(ret=ret, state=state, **pending_kwargs)



# TODO: How to handle corner cases like abcXYZ ?
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Set


# TODO: make this work with the JSON schemas so that we can handles cases such as abcXYZ
def convert_raw_resource_to_present(
    hub,
    raw_resource: Dict[str, Any],
    resource_type_name: str = None,
    exclude_properties_from_transformation: List[str] = None,
) -> Dict[str, Any]:
    return _convert_resource(
        raw_resource,
        hub.tool.gcp.case.snake,
        resource_type_name,
        exclude_properties_from_transformation,
    )


def convert_present_resource_to_raw(
    hub,
    present_resource: Dict[str, Any],
    resource_type_name: str = None,
    exclude_properties_from_transformation: List[str] = None,
) -> Dict[str, Any]:
    return _convert_resource(
        present_resource,
        hub.tool.gcp.case.camel,
        resource_type_name,
        exclude_properties_from_transformation,
    )


def _convert_resource(
    resource,
    key_transformer: Callable[[Any], Any],
    resource_type_name: str = None,
    exclude_properties_from_transformation: List[str] = None,
):
    if resource is None:
        return None

    if exclude_properties_from_transformation is None:
        exclude_properties_from_transformation = []

    if isinstance(resource, list):
        return list(
            _convert_resource(
                v,
                key_transformer,
                resource_type_name,
                exclude_properties_from_transformation,
            )
            for v in resource
        )
    elif isinstance(resource, set):
        return {
            _convert_resource(
                v,
                key_transformer,
                resource_type_name,
                exclude_properties_from_transformation,
            )
            for v in resource
        }
    elif isinstance(resource, dict):
        return {
            k
            if k in exclude_properties_from_transformation
            else key_transformer(k): _convert_resource(
                v,
                key_transformer,
                resource_type_name,
                exclude_properties_from_transformation,
            )
            for k, v in resource.items()
        }
    else:
        return resource


def convert_raw_properties_to_present(
    hub, raw_properties: Set, resource_type_name: str = None
) -> Set:
    if not raw_properties:
        return set()

    properties_translated = set()

    for value in raw_properties:
        properties_translated.add(hub.tool.gcp.case.snake(value))

    return properties_translated


def convert_present_properties_to_raw(
    hub, present_properties: Set, resource_type_name: str = None
) -> Set:
    if not present_properties:
        return set()

    properties_translated = set()

    for value in present_properties:
        properties_translated.add(hub.tool.gcp.case.camel(value))

    return properties_translated



from typing import Any
from typing import Dict
from typing import List


async def update_peerings(
    hub,
    ctx,
    resource_id: str,
    current_peerings: List[Dict[str, Any]],
    new_peerings: List[Dict[str, Any]],
) -> Dict[str, Any]:
    result = {"result": False, "comment": []}

    current_peerings = current_peerings if current_peerings is not None else []
    new_peerings = new_peerings if new_peerings is not None else current_peerings

    peerings_map = {peering.get("name"): peering for peering in current_peerings}
    peerings_to_add = []
    peerings_to_update = []

    for peering in new_peerings:
        if peering.get("name") not in peerings_map:
            peerings_to_add.append(peering)
        else:
            element = peerings_map.pop(peering.get("name"))
            needs_update: bool = (
                element.get("exchange_subnet_routes")
                != peering.get("exchange_subnet_routes")
                or element.get("export_custom_routes")
                != peering.get("export_custom_routes")
                or element.get("export_subnet_routes_with_public_ip")
                != peering.get("export_subnet_routes_with_public_ip")
                or element.get("import_custom_routes")
                != peering.get("import_custom_routes")
                or element.get("import_subnet_routes_with_public_ip")
                != peering.get("import_subnet_routes_with_public_ip")
                or element.get("name") != peering.get("name")
                or element.get("network") != peering.get("network")
                or element.get("peer_mtu") != peering.get("peer_mtu")
                or element.get("stack_type") != peering.get("stack_type")
            )

            if needs_update:
                peerings_to_update.append(peering)

    for peering in peerings_map.values():
        remove_peering_request_body = {"name": peering.get("name")}

        remove_ret = await hub.exec.gcp_api.client.compute.network.removePeering(
            ctx, resource_id=resource_id, body=remove_peering_request_body
        )

        r = await hub.tool.gcp.operation_utils.await_operation_completion(
            ctx, remove_ret, "compute.network", "compute.global_operation"
        )
        if not r["result"]:
            result["comment"] += r["comment"]
            return result

    for peering in peerings_to_update:
        update_peering_request_body = build_peering_request_body(peering)

        update_ret = await hub.exec.gcp_api.client.compute.network.updatePeering(
            ctx, resource_id=resource_id, body=update_peering_request_body
        )

        r = await hub.tool.gcp.operation_utils.await_operation_completion(
            ctx, update_ret, "compute.network", "compute.global_operation"
        )
        if not r["result"]:
            result["comment"] += r["comment"]
            return result

    r = await hub.tool.gcp.compute.network.add_peerings(
        ctx, resource_id, peerings_to_add
    )

    if not r["result"]:
        result["comment"] += r["comment"]
        return result

    result["result"] = True
    return result


async def add_peerings(
    hub, ctx, resource_id: str, peerings_to_add: Dict[str, Any]
) -> Dict[str, Any]:
    result = {"result": False, "comment": []}

    for peering in peerings_to_add:
        add_peering_request_body = build_peering_request_body(peering)

        add_ret = await hub.exec.gcp_api.client.compute.network.addPeering(
            ctx, resource_id=resource_id, body=add_peering_request_body
        )

        r = await hub.tool.gcp.operation_utils.await_operation_completion(
            ctx, add_ret, "compute.network", "compute.global_operation"
        )
        if not r["result"]:
            result["comment"] += r["comment"]
            return result

    result["result"] = True
    return result


def build_peering_request_body(
    peering: Dict[str, Dict[str, Any]]
) -> Dict[str, Dict[str, Any]]:
    return {
        "network_peering": {
            "exchange_subnet_routes": peering.get("exchange_subnet_routes"),
            "export_custom_routes": peering.get("export_custom_routes"),
            "export_subnet_routes_with_public_ip": peering.get(
                "export_subnet_routes_with_public_ip"
            ),
            "import_custom_routes": peering.get("import_custom_routes"),
            "import_subnet_routes_with_public_ip": peering.get(
                "import_subnet_routes_with_public_ip"
            ),
            "name": peering.get("name"),
            "network": peering.get("network"),
            "peer_mtu": peering.get("peer_mtu"),
            "stack_type": peering.get("stack_type"),
        }
    }



"""Methods dealing with Cloud KMS IAM policies."""
from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import List


@dataclass
class AuditLogConfig:
    logType: str
    exemptedMembers: List[str]


@dataclass
class AuditConfig:
    service: str
    auditLogConfigs: List[AuditLogConfig]


@dataclass
class Expr:
    expression: str
    title: str
    description: str = None
    location: str = None


@dataclass
class Binding:
    role: str
    members: List[str]
    condition: Expr = None


@dataclass
class Policy:
    version: int
    bindings: List[Binding]
    auditConfigs: List[AuditConfig] = None
    etag: str = None


async def set_iam_policy(
    hub, ctx, sub, resource: str, policy: Policy, update_mask: str
) -> Dict[str, Any]:

    ret = await sub.setIamPolicy(
        ctx, resource=resource, body={"policy": policy, "updateMask": update_mask}
    )
    return ret


async def get_iam_policy(
    hub, ctx, sub, resource: str, requested_version
) -> Dict[str, Any]:
    ret = await sub.getIamPolicy(
        ctx, resource=resource, __options_requestedPolicyVersion=requested_version
    )
    return ret


# policy_ret = await hub.tool.gcp.policy.get_iam_policy(
#      ctx,
#      hub.exec.gcp_api.client.cloudkms.projects.locations.key_rings,
#      resource_id, 3)

# policy = Policy(version=3,
#                 bindings=[
#                     Binding(role="roles/cloudkms.cryptoKeyDecrypter",
#                             members=["user:dantovska@vmware.com"],
#                             condition=Expr(title="expirable condition",
#                                            expression="request.time < timestamp('2023-10-01T00:00:00.000Z')"))])
# new_policy_ret = await hub.tool.gcp.policy.set_iam_policy(
#     ctx,
#     hub.exec.gcp_api.client.cloudkms.projects.locations.key_rings,
#     resource_id,
#     dataclasses.asdict(policy),
#     None)



"""State module for managing Cloud Key Management Service crypto keys."""
from copy import deepcopy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    crypto_key_version_id: str = None,
    project_id: str = None,
    location_id: str = None,
    key_ring_id: str = None,
    crypto_key_id: str = None,
    key_state: str = None,
    protection_level: str = None,
    algorithm: str = None,
    attestation: make_dataclass(
        "KeyOperationAttestation",
        [
            ("format", str, field(default=None)),
            ("content", str, field(default=None)),
            (
                "cert_chains",
                make_dataclass(
                    "CertificateChains",
                    [
                        ("cavium_certs", List[str], field(default=None)),
                        ("google_card_certs", List[str], field(default=None)),
                        ("google_partition_certs", List[str], field(default=None)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
    create_time: str = None,
    generate_time: str = None,
    destroy_time: str = None,
    destroy_event_time: str = None,
    import_job: str = None,
    import_time: str = None,
    import_failure_reason: str = None,
    external_protection_level_options: make_dataclass(
        "ExternalProtectionLevelOptions",
        [
            ("external_key_uri", str, field(default=None)),
            ("ekm_connection_key_path", str, field(default=None)),
        ],
    ) = None,
    reimport_eligible: bool = None,
    resource_id: str = None,
    key_material: str = None,
) -> Dict[str, Any]:
    """Create or update a `CryptoKeyVersion`_ within a `CryptoKey`_.

    Args:
        name(str):
            Idem name

        crypto_key_version_id(str, Optional):
            Output only. Set by the service.

        project_id(str, Optional):
            Project Id of the new crypto key version.

        location_id(str, Optional):
            Location Id of the new crypto key version .

        key_ring_id(str, Optional):
            Keyring Id of the new crypto key version.

        crypto_key_id(str, Optional):
            Cryptokey Id of the new crypto key version.

        key_state(str, Optional):
            The current state of the `CryptoKeyVersion`_.

        protection_level(str, Optional):
            Output only. The `ProtectionLevel`_ describing how crypto operations are performed with this `CryptoKeyVersion`_.

        algorithm(str, Optional):
            Output only. The `CryptoKeyVersionAlgorithm`_ that this `CryptoKeyVersion`_ supports.

        attestation(Dict[str, Any], Optional):
            Output only. Statement that was generated and signed by the HSM at key creation time. Use this statement to
            verify attributes of the key as stored on the HSM, independently of Google. Only provided for key versions
            with `protectionLevel`_ `HSM`_.

        create_time(str, Optional):
            Output only. The time at which this `CryptoKeyVersion`_ was created.

            A timestamp in RFC3339 UTC "Zulu" format, with nanosecond resolution and up to nine fractional digits.
            Examples: "2014-10-02T15:01:23Z" and "2014-10-02T15:01:23.045123456Z".

        generate_time(str, Optional):
            Output only. The time this `CryptoKeyVersion`_'s key material was generated.

            A timestamp in RFC3339 UTC "Zulu" format, with nanosecond resolution and up to nine fractional digits.
            Examples: "2014-10-02T15:01:23Z" and "2014-10-02T15:01:23.045123456Z".

        destroy_time(str, Optional):
            Output only. The time this `CryptoKeyVersion`_'s key material is scheduled for destruction. Only present if
            `state`_ is `DESTROY_SCHEDULED`_.

            A timestamp in RFC3339 UTC "Zulu" format, with nanosecond resolution and up to nine fractional digits.
            Examples: "2014-10-02T15:01:23Z" and "2014-10-02T15:01:23.045123456Z".

        destroy_event_time(str, Optional):
            Output only. The time this `CryptoKeyVersion`_'s key material was destroyed. Only present if `state`_ is `DESTROYED`_.

            A timestamp in RFC3339 UTC "Zulu" format, with nanosecond resolution and up to nine fractional digits.
            Examples: "2014-10-02T15:01:23Z" and "2014-10-02T15:01:23.045123456Z".

        import_job(str, Optional):
            Output only. The name of the `ImportJob`_ used in the most recent import of this `CryptoKeyVersion`_. Only
            present if the underlying key material was imported.

        import_time(str, Optional):
            Output only. The time at which this `CryptoKeyVersion`_'s key material was most recently imported.

            A timestamp in RFC3339 UTC "Zulu" format, with nanosecond resolution and up to nine fractional digits.
            Examples: "2014-10-02T15:01:23Z" and "2014-10-02T15:01:23.045123456Z".

        import_failure_reason(str, Optional):
            Output only. The root cause of the most recent import failure. Only present if state is `IMPORT_FAILED`_.

        external_protection_level_options(Dict[str, Any], Optional):
            `ExternalProtectionLevelOptions`_ stores a group of additional fields for configuring a `CryptoKeyVersion`_
             that are specific to the `EXTERNAL`_ protection level and `EXTERNAL_VPC`_ protection levels.

        reimport_eligible(bool, Optional):
            Output only. Whether or not this key version is eligible for reimport, by being specified as a target in
            `ImportCryptoKeyVersionRequest.crypto_key_version`_.

        resource_id(str, Optional): Idem resource id. Formatted as

            `projects/{project_id}/locations/{location_id}/keyRings/{key_ring_id}/cryptoKeys/{crypto_key_id}/cryptoKeyVersions/{crypto_key_version_id}`

        key_material(str, Optional): Base64 encoded key material. If this parameter is present will be attempted `import`_
            of the key material in the `CryptoKeyVersion`_ specified by the resource_id or in a new `CryptoKeyVersion`_
            if resource_id is missing. `import`_ requires also project_id, location_id, key_ring_id, crypto_key_id,
            algorithm and import_job parameters to be provided.

    .. _CryptoKey: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.cryptoKeys#CryptoKey
    .. _CryptoKeyVersion: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.cryptoKeys.cryptoKeyVersions#CryptoKeyVersion
    .. _ProtectionLevel: https://cloud.google.com/kms/docs/reference/rest/v1/ProtectionLevel
    .. _CryptoKeyVersionAlgorithm: https://cloud.google.com/kms/docs/reference/rest/v1/CryptoKeyVersionAlgorithm
    .. _protectionLevel: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.cryptoKeys.cryptoKeyVersions#CryptoKeyVersion.FIELDS.protection_level
    .. _HSM: https://cloud.google.com/kms/docs/reference/rest/v1/ProtectionLevel#ENUM_VALUES.HSM
    .. _state: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.cryptoKeys.cryptoKeyVersions#CryptoKeyVersion.FIELDS.state
    .. _DESTROYED: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.cryptoKeys.cryptoKeyVersions#CryptoKeyVersion.CryptoKeyVersionState.ENUM_VALUES.DESTROYED
    .. _DESTROY_SCHEDULED: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.cryptoKeys.cryptoKeyVersions#CryptoKeyVersion.CryptoKeyVersionState.ENUM_VALUES.DESTROY_SCHEDULED
    .. _ImportJob: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.importJobs#ImportJob
    .. _IMPORT_FAILED: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.cryptoKeys.cryptoKeyVersions#CryptoKeyVersion.CryptoKeyVersionState.ENUM_VALUES.IMPORT_FAILED
    .. _ExternalProtectionLevelOptions: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.cryptoKeys.cryptoKeyVersions#CryptoKeyVersion.ExternalProtectionLevelOptions
    .. _EXTERNAL: https://cloud.google.com/kms/docs/reference/rest/v1/ProtectionLevel#ENUM_VALUES.EXTERNAL
    .. _EXTERNAL_VPC: https://cloud.google.com/kms/docs/reference/rest/v1/ProtectionLevel#ENUM_VALUES.EXTERNAL_VPC
    .. _ImportCryptoKeyVersionRequest.crypto_key_version: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.cryptoKeys.cryptoKeyVersions/import#body.request_body.FIELDS.crypto_key_version
    .. _import: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.cryptoKeys.cryptoKeyVersions/import

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

        crypto_key_test:
          gcp.cloudkms.crypto_key.present:
            - project_id: tango-gcp
            - location_id: us-east1
            - key_ring_id: idem-gcp-1
            - crypto_key_id: key-2

        crypto_key_version_test:
          gcp.cloudkms.crypto_key_version.present:
            - key_state: ENABLED
            - project_id: tango-gcp
            - location_id: us-east1
            - key_ring_id: idem-gcp-1
            - crypto_key_id: "${gcp.cloudkms.crypto_key:crypto_key_test:crypto_key_id}"

          # Update crypto key primary version with the one managed above
          gcp.cloudkms.crypto_key.present:
            - primary:
                name: "${gcp.cloudkms.crypto_key_version:crypto_key_version_test:resource_id}"
            - project_id: tango-gcp
            - location_id: us-east1
            - key_ring_id:  idem-gcp-1
            - crypto_key_id: "${gcp.cloudkms.crypto_key:crypto_key_test:crypto_key_id}"
    """
    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }

    create_if_missing = False
    if not resource_id:
        if project_id and location_id and key_ring_id and crypto_key_id:
            if crypto_key_version_id:
                resource_id = hub.tool.gcp.resource_prop_utils.construct_resource_id(
                    "cloudkms.projects.locations.key_rings.crypto_keys.crypto_key_versions",
                    {
                        "project_id": project_id,
                        "location_id": location_id,
                        "key_ring_id": key_ring_id,
                        "crypto_key_id": crypto_key_id,
                        "crypto_key_version_id": crypto_key_version_id,
                    },
                )
            else:
                create_if_missing = True
        else:
            result["result"] = False
            result["comment"].append(
                "When creating new resource crypto_key_version_id, project_id, location_id, key_ring_id and crypto_key_id parameters are required!"
            )
            return result
    else:
        els = hub.tool.gcp.resource_prop_utils.get_elements_from_resource_id(
            "cloudkms.projects.locations.key_rings.crypto_keys.crypto_key_versions",
            resource_id,
        )
        if (
            project_id
            and location_id
            and key_ring_id
            and crypto_key_id
            and crypto_key_version_id
        ):
            if (
                els.get("project_id") != project_id
                or els.get("location_id") != location_id
                or els.get("key_ring_id") != key_ring_id
                or els.get("crypto_key_id") != crypto_key_id
                or els.get("crypto_key_version_id") != str(crypto_key_version_id)
            ):
                result["result"] = False
                result["comment"].append(
                    hub.tool.gcp.comment_utils.non_updatable_properties_comment(
                        "gcp.cloudkms.crypto_key",
                        resource_id,
                        {
                            "project_id",
                            "location_id",
                            "key_ring_id",
                            "crypto_key_id",
                            "crypto_key_version_id",
                        },
                    )
                )
                return result
        else:
            project_id = els.get("project_id")
            location_id = els.get("location_id")
            key_ring_id = els.get("key_ring_id")
            crypto_key_id = els.get("crypto_key_id")
            crypto_key_version_id = els.get("crypto_key_version_id")

    is_create = False
    if resource_id:
        old_get_ret = await hub.exec.gcp.cloudkms.crypto_key_version.get(
            ctx, resource_id=resource_id
        )
        if not old_get_ret["result"] or not old_get_ret["ret"]:
            if not create_if_missing:
                result["result"] = False
                result["comment"] += old_get_ret["comment"]
                return result
            is_create = True
        else:
            result["old_state"] = {
                "name": name,
                "project_id": project_id,
                "location_id": location_id,
                "key_ring_id": key_ring_id,
                "crypto_key_id": crypto_key_id,
                "crypto_key_version_id": crypto_key_version_id,
                **hub.tool.gcp.cloudkms.crypto_key_version_utils.to_state(
                    old_get_ret["ret"]
                ),
            }
            if (
                result["old_state"].get("key_state") != "PENDING_IMPORT"
                and key_state == "PENDING_IMPORT"
                and ctx.get("rerun_data")
                and ctx.get("rerun_data").get("imported_key_material")
            ):
                key_state = result["old_state"].get("key_state")

    else:
        is_create = create_if_missing

    resource_body = {
        "state": key_state,
        "external_protection_level_options": external_protection_level_options,
    }
    resource_body = {k: v for (k, v) in resource_body.items() if v is not None}

    import_key_ret = (
        await hub.tool.gcp.cloudkms.crypto_key_version_utils.get_import_key_if_eligible(
            ctx,
            key_material is not None,
            algorithm is not None,
            import_job,
            result["old_state"].get("key_state")
            if not is_create and "old_state" in result
            else None,
        )
    )
    if not import_key_ret["result"] and not (
        ctx.get("rerun_data") and ctx.get("rerun_data").get("imported_key_material")
    ):
        result["result"] = False
        result["comment"] += import_key_ret["comment"]
        return result

    if ctx["test"]:
        if import_key_ret["ret"]:
            if is_create:
                result["comment"] += (
                    "Key material will be imported in new gcp.cloudkms.crypto_key_version",
                )
            else:
                result["comment"] += (
                    f"Key material will be re-imported in gcp.cloudkms.crypto_key_version '{resource_id}'",
                )
            return result
        if is_create:
            result["comment"].append(
                hub.tool.gcp.comment_utils.would_create_comment(
                    "gcp.cloudkms.crypto_key_version", resource_id
                )
            )
            result["new_state"] = {
                "resource_id": resource_id,
                "name": name,
                "project_id": project_id,
                "location_id": location_id,
                "key_ring_id": key_ring_id,
                "crypto_key_id": crypto_key_id,
                "crypto_key_version_id": crypto_key_version_id,
                **resource_body,
            }
        else:
            result["new_state"] = {
                **deepcopy(result["old_state"]),
                **{
                    "key_state" if k == "state" else k: v
                    for k, v in resource_body.items()
                },
            }
            update_mask = hub.tool.gcp.cloudkms.patch.calc_update_mask(
                resource_body,
                {
                    "state" if k == "key_state" else k: v
                    for k, v in result["old_state"].items()
                },
            )
            if update_mask:
                result["comment"].append(
                    hub.tool.gcp.comment_utils.would_update_comment(
                        "gcp.cloudkms.crypto_key_version", resource_id
                    )
                )
            else:
                result["comment"].append(
                    hub.tool.gcp.comment_utils.already_exists_comment(
                        "gcp.cloudkms.crypto_key_version", resource_id
                    )
                )
        return result

    parent = hub.tool.gcp.resource_prop_utils.construct_resource_id(
        "cloudkms.projects.locations.key_rings.crypto_keys",
        {
            "project_id": project_id,
            "location_id": location_id,
            "key_ring_id": key_ring_id,
            "crypto_key_id": crypto_key_id,
        },
    )
    if import_key_ret["result"] and import_key_ret["ret"]:
        import_ret = await hub.exec.gcp.cloudkms.crypto_key_version["import"](
            ctx,
            parent=parent,
            import_job=import_job,
            import_job_pub_key=import_key_ret["ret"],
            algorithm=algorithm,
            key_material=key_material,
            crypto_key_version=resource_id,
        )
        if not import_ret["result"]:
            result["result"] = False
            result["comment"] += import_ret["comment"]
            return result

        result["new_state"] = {
            "name": name,
            "project_id": project_id,
            "location_id": location_id,
            "key_ring_id": key_ring_id,
            "crypto_key_id": crypto_key_id,
            "crypto_key_version_id": crypto_key_version_id,
            **hub.tool.gcp.cloudkms.crypto_key_version_utils.to_state(
                import_ret["ret"]
            ),
        }
        result["rerun_data"] = {"imported_key_material": True}
        result["comment"].append(
            f"Key material {'imported' if is_create else 're-imported'} in gcp.cloudkms.crypto_key_version {result['new_state']['resource_id']}"
        )
        return result

    if is_create:
        create_ret = await hub.exec.gcp_api.client.cloudkms.projects.locations.key_rings.crypto_keys.crypto_key_versions.create(
            ctx,
            parent=parent,
            body=resource_body,
        )
        if not create_ret["result"]:
            result["result"] = False
            result["comment"] += create_ret["comment"]
            return result
        result["new_state"] = {
            "name": name,
            "project_id": project_id,
            "location_id": location_id,
            "key_ring_id": key_ring_id,
            "crypto_key_id": crypto_key_id,
            "crypto_key_version_id": crypto_key_version_id,
            **hub.tool.gcp.cloudkms.crypto_key_version_utils.to_state(
                create_ret["ret"]
            ),
        }
        result["comment"].append(
            hub.tool.gcp.comment_utils.create_comment(
                "gcp.cloudkms.crypto_key_version", result["new_state"]["resource_id"]
            )
        )
    else:
        update_mask = hub.tool.gcp.cloudkms.patch.calc_update_mask(
            resource_body,
            {
                "state" if k == "key_state" else k: v
                for k, v in result["old_state"].items()
            },
        )
        if (
            update_mask
            and result["old_state"].get("key_state") == "DESTROY_SCHEDULED"
            and key_state == "ENABLED"
        ):
            restore_ret = await hub.exec.gcp_api.client.cloudkms.projects.locations.key_rings.crypto_keys.crypto_key_versions.restore(
                ctx, name_=resource_id
            )
            if not restore_ret["result"]:
                result["result"] = False
                result["comment"] += restore_ret["comment"]
                return result
        if update_mask:
            update_ret = await hub.exec.gcp_api.client.cloudkms.projects.locations.key_rings.crypto_keys.crypto_key_versions.patch(
                ctx, name_=resource_id, updateMask=update_mask, body=resource_body
            )
            if not update_ret["result"]:
                result["result"] = False
                result["comment"] += update_ret["comment"]
                return result

            result["comment"].append(
                hub.tool.gcp.comment_utils.update_comment(
                    "gcp.cloudkms.crypto_key_version", resource_id
                )
            )
            result["new_state"] = {
                "name": name,
                "project_id": project_id,
                "location_id": location_id,
                "key_ring_id": key_ring_id,
                "crypto_key_id": crypto_key_id,
                "crypto_key_version_id": crypto_key_version_id,
                **hub.tool.gcp.cloudkms.crypto_key_version_utils.to_state(
                    update_ret["ret"]
                ),
            }
        else:
            result["comment"].append(
                hub.tool.gcp.comment_utils.already_exists_comment(
                    "gcp.cloudkms.crypto_key_version", resource_id
                )
            )
            result["new_state"] = deepcopy(result["old_state"])

    return result


async def absent(
    hub,
    ctx,
    name: str,
    crypto_key_version_id: str = None,
    project_id: str = None,
    location_id: str = None,
    key_ring_id: str = None,
    crypto_key_id: str = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    """Destroy crypto key version.

    After this operation the key material will no longer be stored. This version may only become `ENABLED` again if this
    version is `reimportEligible`_ and the original key material is reimported with a call to
    `KeyManagementService.ImportCryptoKeyVersion`_. Should provide either resource_id or all other *_id parameters.

    .. _KeyManagementService.ImportCryptoKeyVersion: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.cryptoKeys.cryptoKeyVersions/import#google.cloud.kms.v1.KeyManagementService.ImportCryptoKeyVersion
    .. _reimportEligible: https://cloud.google.com/kms/docs/reference/rest/v1/projects.locations.keyRings.cryptoKeys.cryptoKeyVersions#CryptoKeyVersion.FIELDS.reimport_eligible

    Args:
        name(str):
            Idem name.

        crypto_key_version_id(str, Optional):
            Crypto key version name used to generate resource_id if it is not provided.

        project_id(str, Optional):
            Project Id of the new crypto key version.

        location_id(str, Optional):
            Location Id of the new crypto key version .

        key_ring_id(str, Optional):
            Keyring Id of the new crypto key version.

        crypto_key_id(str, Optional):
            Cryptokey Id of the new crypto key version.

        resource_id(str, Optional): Idem resource id. Formatted as

            `projects/{project_id}/locations/{location_id}/keyRings/{key_ring_id}/cryptoKeys/{crypto_key_id}/cryptoKeyVersions/{crypto_key_version_id}`
    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            {% set project_id = 'tango-gcp' %}
            {% set location_id = 'us-east1' %}
            {% set key_ring = 'key-ring' %}
            {% set crypto_key = 'crypto-key' %}
            {% set crypto_key_version = 'crypto-key-version' %}
            resource_is_absent:
              gcp.cloudkms.crypto_key_version.absent:
                - resource_id: projects/{{project_id}}/locations/{{location_id}}/keyRings/{{key_ring}}/cryptoKeys/{{crypto_key}}/cryptoKeyVersions/{{crypto_key_version}}
    """
    result = {
        "comment": [],
        "old_state": ctx.get("old_state"),
        "new_state": None,
        "name": name,
        "result": True,
    }

    if not resource_id:
        resource_id = (ctx.old_state or {}).get("resource_id")
        if (
            not resource_id
            and project_id
            and location_id
            and key_ring_id
            and crypto_key_id
            and crypto_key_version_id
        ):
            resource_id = hub.tool.gcp.resource_prop_utils.construct_resource_id(
                "cloudkms.projects.locations.key_rings.crypto_keys.crypto_key_versions",
                {
                    "project_id": project_id,
                    "location_id": location_id,
                    "key_ring_id": key_ring_id,
                    "crypto_key_id": crypto_key_id,
                    "crypto_key_version_id": crypto_key_version_id,
                },
            )
        else:
            result["result"] = False
            result["comment"] += ["Can not determine resource_id."]
            return result

    if ctx.test:
        result["comment"].append(
            hub.tool.gcp.comment_utils.would_delete_comment(
                "gcp.cloudkms.crypto_key_version", resource_id
            )
        )
        return result

    get_ret = await hub.exec.gcp.cloudkms.crypto_key_version.get(
        ctx, resource_id=resource_id
    )
    if not get_ret["result"]:
        result["comment"] += get_ret["comment"]
        return result
    elif not get_ret["ret"] or get_ret["ret"]["state"] in [
        "DESTROYED",
        "DESTROY_SCHEDULED",
    ]:
        result["comment"].append(
            hub.tool.gcp.comment_utils.already_absent_comment(
                "gcp.cloudkms.crypto_key_version", resource_id
            )
        )
        return result

    destroy_ret = await hub.exec.gcp_api.client.cloudkms.projects.locations.key_rings.crypto_keys.crypto_key_versions.destroy(
        ctx, name_=resource_id
    )

    if destroy_ret["result"]:
        result["comment"].append(
            hub.tool.gcp.comment_utils.delete_comment(
                "gcp.cloudkms.crypto_key_version", resource_id
            )
        )
    else:
        result["comment"] += destroy_ret["comment"]
        result["result"] = False

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Retrieve the list of available crypto key versions.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe gcp.cloudkms.crypto_key_version
    """
    result = {}

    locations = await hub.exec.gcp.cloudkms.location.list(
        ctx, project=ctx.acct.project_id
    )
    if not locations["result"]:
        hub.log.warning(
            f"Could not list gcp.cloudkms.location in {ctx.acct.project_id} {locations['comment']}"
        )
        return {}

    for location in locations["ret"]:
        key_rings = await hub.exec.gcp.cloudkms.key_ring.list(
            ctx, location=location["resource_id"]
        )
        if not key_rings["result"]:
            hub.log.warning(
                f"Could not list gcp.cloudkms.key_ring in {location['location_id']} {key_rings['comment']}"
            )
        else:
            for key_ring in key_rings["ret"]:
                crypto_keys = await hub.exec.gcp.cloudkms.crypto_key.list(
                    ctx, key_ring=key_ring["resource_id"]
                )
                if not crypto_keys["result"]:
                    hub.log.warning(
                        f"Could not describe gcp.cloudkms.crypto_key in {key_ring['resource_id']} {key_rings['comment']}"
                    )
                else:
                    for crypto_key in crypto_keys["ret"]:
                        crypto_key_versions = (
                            await hub.exec.gcp.cloudkms.crypto_key_version.list(
                                ctx, crypto_key=crypto_key["resource_id"]
                            )
                        )
                        if not crypto_keys["result"]:
                            hub.log.warning(
                                f"Could not describe gcp.cloudkms.crypto_key in {key_ring['resource_id']} {key_rings['comment']}"
                            )
                        else:
                            for crypto_key_version in crypto_key_versions["ret"]:
                                resource_id = crypto_key_version["resource_id"]
                                result[resource_id] = {
                                    "gcp.cloudkms.crypto_key_version.present": [
                                        {parameter_key: parameter_value}
                                        if parameter_key != "state"
                                        else {"key_state": parameter_value}
                                        for parameter_key, parameter_value in crypto_key_version.items()
                                    ]
                                }
                                els = hub.tool.gcp.resource_prop_utils.get_elements_from_resource_id(
                                    "cloudkms.projects.locations.key_rings.crypto_keys.crypto_key_versions",
                                    resource_id,
                                )
                                p = result[resource_id][
                                    "gcp.cloudkms.crypto_key_version.present"
                                ]
                                p.append({"project_id": els["project_id"]})
                                p.append({"location_id": els["location_id"]})
                                p.append({"key_ring_id": els["key_ring_id"]})
                                p.append({"crypto_key_id": els["crypto_key_id"]})
                                p.append(
                                    {
                                        "crypto_key_version_id": els[
                                            "crypto_key_version_id"
                                        ]
                                    }
                                )

    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    return hub.tool.gcp.utils.is_pending(ret=ret, state=state, **pending_kwargs)



from idem_gcp.tool.gcp.generate.exec_context import ExecutionContext
from idem_gcp.tool.gcp.generate.exec_param import ExecParam


async def should_operate(hub, execution_context: ExecutionContext) -> bool:
    return (
        execution_context.get_exec_param_value(ExecParam.GET_ITEMS)
        and "list" in execution_context.method_name.lower()
        and execution_context.response.get("ret") is not None
        and "items" in execution_context.response.get("ret")
    )


async def operate(hub, execution_context: ExecutionContext) -> None:
    response = execution_context.response
    execution_context.response = {**response, "ret": response.get("ret").get("items")}



"""State module for managing Resource Policies."""
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List


__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str = None,
    project: str = None,
    region: str = None,
    resource_id: str = None,
    description: str = None,
    snapshot_schedule_policy: make_dataclass(
        "ResourcePolicySnapshotSchedulePolicy",
        [
            (
                "schedule",
                make_dataclass(
                    "ResourcePolicySnapshotSchedulePolicySchedule",
                    [
                        (
                            "daily_schedule",
                            make_dataclass(
                                "ResourcePolicyDailyCycle",
                                [
                                    ("days_in_cycle", int, field(default=0)),
                                    ("start_time", str, field(default=None)),
                                ],
                            ),
                            field(default=None),
                        ),
                        (
                            "weekly_schedule",
                            make_dataclass(
                                "ResourcePolicyWeeklyCycle",
                                [
                                    (
                                        "day_of_weeks",
                                        List[
                                            make_dataclass(
                                                "ResourcePolicyWeeklyCycleDayOfWeek",
                                                [
                                                    (
                                                        "start_time",
                                                        str,
                                                        field(default=None),
                                                    ),
                                                    ("day", str, field(default=None)),
                                                ],
                                            )
                                        ],
                                        field(default=None),
                                    ),
                                ],
                            ),
                            field(default=None),
                        ),
                        (
                            "hourly_schedule",
                            make_dataclass(
                                "ResourcePolicyHourlyCycle",
                                [
                                    ("hours_in_cycle", int, field(default=0)),
                                    ("start_time", str, field(default=None)),
                                ],
                            ),
                            field(default=None),
                        ),
                    ],
                ),
                field(default=None),
            ),
            (
                "retention_policy",
                make_dataclass(
                    "ResourcePolicySnapshotSchedulePolicyRetentionPolicy",
                    [
                        ("max_retention_days", int, field(default=0)),
                        ("on_source_disk_delete", str, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "snapshotProperties",
                make_dataclass(
                    "ResourcePolicySnapshotSchedulePolicySnapshotProperties",
                    [
                        ("guest_flush", bool, field(default=False)),
                        ("chain_name", str, field(default=None)),
                        ("storage_locations", List[str], field(default=None)),
                        ("labels", Dict[str, str], field(default=None)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
    group_placement_policy: make_dataclass(
        "ResourcePolicyGroupPlacementPolicy",
        [
            ("vm_count", int, field(default=0)),
            ("collocation", str, field(default=None)),
            ("availability_domain_count", int, field(default=0)),
        ],
    ) = None,
    instance_schedule_policy: make_dataclass(
        "ResourcePolicyInstanceSchedulePolicy",
        [
            ("start_time", str, field(default=None)),
            ("time_zone", str, field(default=None)),
            ("expiration_time", str, field(default=None)),
            (
                "vm_start_schedule",
                make_dataclass(
                    "ResourcePolicyInstanceSchedulePolicySchedule",
                    [
                        ("schedule", str, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
            (
                "vm_stop_schedule",
                make_dataclass(
                    "ResourcePolicyInstanceSchedulePolicySchedule",
                    [
                        ("schedule", str, field(default=None)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    r"""Creates a new resource policy.

    You can use resource policies to schedule actions for some Compute Engine resources.
    For example, you can use them to schedule persistent disk snapshots.

    Args:
        name(str, Optional):
            An Idem name of the resource.

        project(str, Optional):
            Project ID for this request. Defaults to None.

        region(str, Optional):
            Name of the region for this request. Defaults to None.

        resource_id(str, Optional):
            An identifier of the resource in the provider. Defaults to None.

        description(str, Optional):
            An optional description of this resource. Defaults to None.

        snapshot_schedule_policy(Dict[str, Any], Optional):
            Resource policy for persistent disks for creating snapshots. Defaults to None.
            A snapshot schedule policy specifies when and how frequently snapshots are to be created for
            the target disk. Also specifies how many and how long these scheduled snapshots should be retained.

            * schedule(Dict[str, Any], Optional):
                A Vm Maintenance Policy specifies what kind of infrastructure maintenance we are allowed to perform on
                this VM and when. Schedule that is applied to disks covered by this policy.
                A schedule for disks where the scheduled operations are performed.

                * daily_schedule(Dict[str, Any], Optional):
                    Time window specified for daily operations.

                    * days_in_cycle(int, Optional):
                        Defines a schedule with units measured in days.
                        The value determines how many days pass between the start of each cycle.

                    * start_time(str, Optional):
                        Start time of the window. This must be in UTC format that resolves to one of
                        00:00, 04:00, 08:00, 12:00, 16:00, or 20:00. For example, both 13:00-5 and 08:00 are valid.

                * weekly_schedule(Dict[str, Any], Optional):
                    Time window specified for weekly operations.

                    * day_of_weeks(List[Dict[str, Any]], Optional):
                        Up to 7 intervals/windows, one for each day of the week.

                        * start_time(str, Optional):
                            Time within the window to start the operations.
                            It must be in format \"HH:MM\", where HH : [00-23] and MM : [00-00] GMT.

                        * day(str, Optional):
                            Defines a schedule that runs on specific days of the week. Specify one or more days.
                            The following options are available:
                            MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY, SATURDAY, SUNDAY.

                * hourly_schedule(Dict[str, Any], Optional):
                    Time window specified for hourly operations.

                    * hours_in_cycle(int, Optional):
                        Defines a schedule with units measured in hours.
                        The value determines how many hours pass between the start of each cycle.

                    * start_time(str, Optional):
                        Time within the window to start the operations.
                        It must be in format \"HH:MM\", where HH : [00-23] and MM : [00-00] GMT.

            * retention_policy(Dict[str, Any], Optional):
                Retention policy applied to snapshots created by this resource policy.
                Policy for retention of scheduled snapshots.

                * max_retention_days(int, Optional):
                    Maximum age of the snapshot that is allowed to be kept.

                * on_source_disk_delete(str, Optional):
                    Specifies the behavior to apply to scheduled snapshots when the source disk is deleted.
                    The following options are available:
                    APPLY_RETENTION_POLICY, KEEP_AUTO_SNAPSHOTS, UNSPECIFIED_ON_SOURCE_DISK_DELETE


            * snapshot_properties(Dict[str, Any], Optional):
                Properties with which snapshots are created such as labels, encryption keys.
                Specified snapshot properties for scheduled snapshots created by this policy.

                * guest_flush(bool, Optional):
                    Indication to perform a 'guest aware' snapshot.

                * chain_name(str, Optional):
                    Chain name that the snapshot is created in.

                * storage_locations(List[str, str], Optional):
                    Cloud Storage bucket storage location of the auto snapshot (regional or multi-regional).

                * labels(List[str, str], Optional):
                    Labels to apply to scheduled snapshots. These can be later modified by the setLabels method.
                    Label values may be empty.

        group_placement_policy(Dict[str, Any], Optional):
            Resource policy for instances for placement configuration. Defaults to None.
            A GroupPlacementPolicy specifies resource placement configuration.
            It specifies the failure bucket separation as well as network locality

            * vm_count(int, Optional):
                Number of VMs in this placement group. Google does not recommend that you use this field unless
                you use a compact policy, and you want your policy to work only if it contains this exact number of VMs.

            * collocation(str, Optional):
                Specifies network collocation. The following options are available:
                COLLOCATED, UNSPECIFIED_COLLOCATION

            * availability_domain_count(int, Optional):
                The number of availability domains to spread instances across.
                If two instances are in different availability domain, they are not in the same low latency network.

        instance_schedule_policy(Dict[str, Any], Optional):
            Resource policy for scheduling instance operations. Defaults to None.
            An InstanceSchedulePolicy specifies when and how frequent certain operations are performed on the instance.

            * start_time(str, Optional):
                The start time of the schedule. The timestamp is an RFC3339 string.

            * time_zone(str, Optional):
                Specifies the time zone to be used in interpreting Schedule.schedule.
                The value of this field must be a time zone name from the tz database:
                http://en.wikipedia.org/wiki/Tz_database.

            * expiration_time(str, Optional):
                The expiration time of the schedule. The timestamp is an RFC3339 string.

            * vm_start_schedule(Dict[str, Any], Optional):
                Specifies the schedule for starting instances.

                * schedule(str, Optional):
                    Specifies the frequency for the operation, using the unix-cron format.

            * vm_stop_schedule(Dict[str, Any], Optional):
                Specifies the schedule for stopping instances.

                * schedule(str, Optional):
                    Specifies the frequency for the operation, using the unix-cron format.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            example_resource_name:
              gcp.compute.resource_policy.present:
                - name: value
                - project: value
                - region: value
    """
    result = {
        "result": True,
        "old_state": None,
        "new_state": None,
        "name": name,
        "comment": [],
    }

    # TODO uncomment below line, when implementation is added
    # project = get_project_from_account(ctx, project)

    result["comment"].append(
        "No-op: There is no create/update function for gcp.compute.resource_policy"
    )

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    request_id: str = None,
) -> Dict[str, Any]:
    r"""Deletes the specified resource policy.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            An identifier of the resource in the provider.
            Defaults to None.

        request_id(str, Optional):
            An optional request ID to identify requests.
            Specify a unique request ID so that if you must retry your request,
            the server will know to ignore the request if it has already been completed.
            For example, consider a situation where you make an initial request and the request times out.
            If you make the request again with the same request ID,
            the server can check if original operation with the same request ID was received, and if so,
            will ignore the second request. This prevents clients from accidentally creating duplicate commitments.
            The request ID must be a valid UUID with the exception that zero UUID is not supported
            ( 00000000-0000-0000-0000-000000000000).
            Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              gcp.compute.resource_policy.absent:
                - resource_id: resource_id
    """
    result = {
        "result": True,
        "old_state": ctx.old_state,
        "new_state": None,
        "name": name,
        "comment": [],
    }

    if not resource_id:
        resource_id = (ctx.old_state or {}).get("resource_id")

    if ctx.test:
        result["comment"].append(
            hub.tool.gcp.comment_utils.would_delete_comment(
                "gcp.compute.resource_policy", name
            )
        )
        return result

    if not ctx.get("rerun_data"):
        delete_ret = await hub.exec.gcp_api.client.compute.resource_policy.delete(
            ctx, resource_id=resource_id
        )
        if delete_ret["ret"]:
            if "compute#operation" in delete_ret["ret"].get("kind"):
                result["result"] = False
                result["comment"] += delete_ret["comment"]
                result[
                    "rerun_data"
                ] = hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
                    delete_ret["ret"].get("selfLink"), "compute.zone_operation"
                )
                return result
    else:
        # delete() has been called on some previous iteration
        handle_operation_ret = await hub.tool.gcp.operation_utils.handle_operation(
            ctx, ctx.get("rerun_data"), "compute.resource_policy"
        )
        if not handle_operation_ret["result"]:
            result["result"] = False
            result["comment"] += handle_operation_ret["comment"]
            result["rerun_data"] = handle_operation_ret["rerun_data"]
            return result

        resource_id = handle_operation_ret["resource_id"]

    if not resource_id:
        result["comment"].append(
            hub.tool.gcp.comment_utils.already_absent_comment(
                "gcp.compute.resource_policy", name
            )
        )
        return result

    result["comment"].append(
        hub.tool.gcp.comment_utils.delete_comment("gcp.compute.resource_policy", name)
    )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Retrieves a list of resource policies contained within the specified region.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe gcp.compute.resource_policy
    """
    result = {}

    describe_ret = await hub.exec.gcp.compute.resource_policy.list(
        ctx, project=ctx.acct.project_id
    )

    if not describe_ret["result"]:
        hub.log.debug(f"Could not describe resource policies {describe_ret['comment']}")
        return {}

    for resource in describe_ret["ret"]:
        resource_id = resource.get("resource_id")
        result[resource_id] = {
            "gcp.compute.resource_policy.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    return hub.tool.gcp.utils.is_pending(ret=ret, state=state, **pending_kwargs)



"""Exec module for managing Cloud Key Management Service locations."""
from typing import Any
from typing import Dict


__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    resource_id: str,
):
    """Returns a location by its Idem resource ID.

    Args:
        resource_id(str):
            Idem resource ID. ``projects/{project id}/locations/{location id}``

    Returns:
        Location resource

    Examples:
        .. code-block:: sls

            {% set project_id = 'project-name' %}
            {% set location_id = 'us-east1' %}
            get-location:
                exec.run:
                    - path: gcp.cloudkms.location.get
                    - kwargs:
                          resource_id: projects/{{project_id}}/locations/{{location_id}}
    """
    result = {
        "comment": [],
        "ret": [],
        "result": True,
    }

    location = await hub.exec.gcp_api.client.cloudkms.projects.locations.get(
        ctx, _name=resource_id
    )

    if not location["result"]:
        result["comment"] += location["comment"]
        result["result"] = False
        return result

    result["ret"] = location["ret"]

    if not result["ret"]:
        result["comment"] += (
            hub.tool.gcp.comment_utils.get_empty_comment(
                "gcp.cloudkms.location", resource_id
            ),
        )

    return result


async def list_(hub, ctx, project: str = None, filter: str = None) -> Dict[str, Any]:
    r"""Retrieves the locations for a specific project.

    Args:
        project(str, Optional):
            Project ID for this request. If not provided will use the one configured in `ctx`

        filter(str, Optional):
            A filter to narrow down results to a preferred subset. The filtering language accepts strings like
            "displayName=tokyo", and is documented in more detail in `AIP-160`_.

    .. _AIP-160: https://google.aip.dev/160

    Examples:
        .. code-block:: sls

            list-locations:
                exec.run:
                    - path: gcp.cloudkms.location.list
                    - kwargs:
                        project: project-name
    """
    result = {
        "comment": [],
        "ret": [],
        "result": True,
    }

    if project:
        name = f"projects/{project}"
    else:
        name = f"projects/{ctx.acct.project_id}"

    locations = await hub.exec.gcp_api.client.cloudkms.projects.locations.list(
        ctx, _name=name, filter=filter
    )

    if not locations["result"]:
        result["comment"] += locations["comment"]
        result["result"] = False
        return result

    if locations["ret"].get("items"):
        result["ret"] = locations["ret"]["items"]

    if not result["ret"]:
        result["comment"] += (
            hub.tool.gcp.comment_utils.list_empty_comment(
                "gcp.cloudkms.location", name
            ),
        )

    return result



import time
from typing import Any
from typing import Dict


async def handle_operation(
    hub, ctx, operation_id: str, resource_type: str, wait_until_done: bool = False
) -> Dict[str, Any]:
    result = {
        "comment": [],
        "result": True,
        "rerun_data": None,
        "resource_id": None,
    }

    operation_type = hub.tool.gcp.operation_utils.get_operation_type(operation_id)

    if operation_type is None:
        result["result"] = False
        result["comment"].append(
            f"Cannot determine operation scope (zonal/regional/global) {operation_id}"
        )
        return result

    if operation_type == "compute.zone_operation":
        get_ret = await hub.exec.gcp_api.client.compute.zone_operation.get(
            ctx, resource_id=operation_id
        )
    elif operation_type == "compute.region_operation":
        get_ret = await hub.exec.gcp_api.client.compute.region_operation.get(
            ctx, resource_id=operation_id
        )
    elif operation_type == "compute.global_operation":
        get_ret = await hub.exec.gcp_api.client.compute.global_operation.get(
            ctx, resource_id=operation_id
        )

    if not get_ret["result"] or not get_ret["ret"]:
        result["result"] = False
        result["comment"] = get_ret["comment"]
        return result

    operation = get_ret["ret"]
    if operation["status"] != "DONE":
        if wait_until_done:
            operation = await hub.tool.gcp.operation_utils.wait_for_operation(
                ctx, operation, operation_type
            )
        else:
            result["result"] = False
            result[
                "rerun_data"
            ] = hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
                operation.get("selfLink"), operation_type
            )
            result["comment"] += get_ret["comment"]
            return result

    if operation.get("error"):
        result["result"] = False
        result["comment"] += str(operation.get("error", {}))
        return result

    result["resource_id"] = hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
        operation.get("targetLink"), resource_type
    )

    return result


async def wait_for_operation(hub, ctx, operation, operation_type: str) -> Dict:
    while True:
        if operation_type == "compute.zone_operation":
            op_ret = await hub.exec.gcp_api.client.compute.zone_operation.wait(
                ctx, resource_id=operation["selfLink"]
            )
        elif operation_type == "compute.region_operation":
            op_ret = await hub.exec.gcp_api.client.compute.region_operation.wait(
                ctx, resource_id=operation["selfLink"]
            )
        elif operation_type == "compute.global_operation":
            op_ret = await hub.exec.gcp_api.client.compute.global_operation.wait(
                ctx, resource_id=operation["selfLink"]
            )

        if not op_ret or not op_ret["ret"] or not op_ret["result"]:
            # wait() call timed out?
            break

        operation = op_ret["ret"]
        if operation["status"] == "DONE" or "error" in operation:
            break

        time.sleep(1)

    return operation


def get_operation_type(hub, operation_id) -> str:
    if "/zones/" in operation_id:
        return "compute.zone_operation"
    elif "/regions/" in operation_id:
        return "compute.region_operation"
    elif "/global/" in operation_id:
        return "compute.global_operation"
    else:
        return None


async def await_operation_completion(
    hub, ctx, api_call_ret: Dict[str, Any], resource_type: str, operation_type: str
) -> Dict[str, Any]:
    result = {"result": False, "comment": []}

    if not api_call_ret["result"] and not api_call_ret["ret"]:
        result["comment"] += api_call_ret["comment"]
        return result

    if "compute#operation" in api_call_ret["ret"].get("kind"):
        operation = api_call_ret["ret"]

        operation_id = hub.tool.gcp.resource_prop_utils.parse_link_to_resource_id(
            operation.get("selfLink"), operation_type
        )
        handle_operation_ret = await hub.tool.gcp.operation_utils.handle_operation(
            ctx, operation_id, resource_type, True
        )

        if not handle_operation_ret["result"]:
            result["comment"] += handle_operation_ret["comment"]
            return result

    result["result"] = True
    return result
